//CanvasGrid v0.3

package
{

import mx.core.Container;
import mx.core.EdgeMetrics;
import mx.styles.IStyleClient;

[Style(name="horizontalGap", type="Number", format="Length", inherit="no")]
[Style(name="verticalGap", type="Number", format="Length", inherit="no")]

public class CanvasGrid extends Container
{
	protected var measuredCellHeight:Number = NaN;
	protected var measuredCellWidth:Number = NaN;

	protected var measuredColumnCount:int = -1;
	protected var measuredRowCount:int = -1;

	private var useGreedyMeasurement:Boolean = false;

	private var _cellHeight:Number = NaN;
	public function get cellHeight():Number
	{
		return _cellHeight;
	}
	public function set cellHeight(value:Number):void
	{
		if (_cellHeight != value)
		{
			_cellHeight = value;

			invalidateSize();
			invalidateDisplayList();
		}
	}

	private var _cellWidth:Number = NaN;
	public function get cellWidth():Number
	{
		return _cellWidth;
	}
	public function set cellWidth(value:Number):void
	{
		if (_cellWidth != value)
		{
			_cellWidth = value;

			invalidateSize();
			invalidateDisplayList();
		}
	}

	private var _columnCount:int = -1;
	public function get columnCount():int
	{
		return _columnCount;
	}
	public function set columnCount(value:int):void
	{
		if (_columnCount != value)
		{
			_columnCount = value;

			invalidateSize();
			invalidateDisplayList();
		}
	}

	private var _rowCount:int = -1;
	public function get rowCount():int
	{
		return _rowCount;
	}
	public function set rowCount(value:int):void
	{
		if (_rowCount != value)
		{
			_rowCount = value;

			invalidateSize();
			invalidateDisplayList();
		}
	}

	override protected function measure():void
	{
		super.measure();

		var vm:EdgeMetrics = viewMetricsAndPadding;

		var hGap:Number = getStyle("horizontalGap");
		var vGap:Number = getStyle("verticalGap");

		var i:int;
		var child:ICanvasGridComponent;

		var cc:int = columnCount;
		var rc:int = rowCount;

		if (cc == -1 || rc == -1)
		{
			var maxC:int = 0;
			var maxR:int = 0;

			for (i = 0; i < numChildren; i++)
			{
				child = getChildAt(i) as ICanvasGridComponent;

				if (cc == -1)
					maxC = Math.max(maxC, child.gridX + child.gridWidth);
				if (rc == -1)
					maxR = Math.max(maxR, child.gridY + child.gridHeight);
			}

			if (cc == -1)
				cc = measuredColumnCount = maxC;
			if (rc == -1)
				rc = measuredRowCount = maxR;
		}

		var cw:Number = cellWidth;
		var ch:Number = cellHeight;

		if (isNaN(cw) && !isNaN(explicitWidth))
			cw = measuredCellWidth = Math.round(
					(explicitWidth - vm.left - vm.right
					 - hGap * (cc - 1)) / cc);
		if (isNaN(ch) && !isNaN(explicitHeight))
			ch = measuredCellHeight = Math.round(
					(explicitHeight - vm.top - vm.bottom
					- vGap * (rc - 1)) / rc);

		if (isNaN(cw) || isNaN(ch))
		{
			var maxCW:Number = 0;
			var maxCH:Number = 0;

			for (i = 0; i < numChildren; i++)
			{
				child = getChildAt(i) as ICanvasGridComponent;

				var styleableChild:IStyleClient;
				var usePercentSizing:Boolean;

				if (isNaN(cw))
				{
					var pw:Number = child.getExplicitOrMeasuredWidth();

					if (useGreedyMeasurement)
					{
						styleableChild = child as IStyleClient;
						usePercentSizing = !isNaN(child.percentWidth);

						var cLeft:Number = NaN;
						var cRight:Number = NaN;
						var cHorizontalCenter:Number = NaN;

						if (styleableChild)
						{
							cLeft = styleableChild.getStyle("left");
							cRight = styleableChild.getStyle("right");
							cHorizontalCenter = styleableChild
								.getStyle("horizontalCenter");

							if (usePercentSizing
									&& ((!isNaN(cLeft)
											&& !isNaN(cRight))
										|| (!isNaN(cLeft)
											&& !isNaN(cHorizontalCenter))
										|| (!isNaN(cRight)
											&& !isNaN(cHorizontalCenter))))
								usePercentSizing = false;
						}

						if (usePercentSizing)
							pw = pw * (100 / child.percentWidth);
						else
						{
							if (!isNaN(cLeft) && !isNaN(cRight))
								pw += cLeft + cRight;
							else if (!isNaN(cLeft) && !isNaN(cHorizontalCenter))
								pw += (cLeft - cHorizontalCenter) * 2;
							else if (!isNaN(cRight)
									&& !isNaN(cHorizontalCenter))
								pw += (cRight + cHorizontalCenter) * 2;
							else if (!isNaN(cLeft))
								pw += cLeft;
							else if (!isNaN(cRight))
								pw += cRight;
							else if (!isNaN(cHorizontalCenter))
								pw += Math.abs(cHorizontalCenter) * 2;
						}
					}

					maxCW = Math.max(maxCW, Math.round(pw / child.gridWidth));
				}

				if (isNaN(ch))
				{
					var ph:Number = child.getExplicitOrMeasuredHeight();

					if (useGreedyMeasurement)
					{
						styleableChild = child as IStyleClient;
						usePercentSizing = !isNaN(child.percentHeight);

						var cTop:Number = NaN;
						var cBottom:Number = NaN;
						var cVerticalCenter:Number = NaN;

						if (styleableChild)
						{
							cTop = styleableChild.getStyle("top");
							cBottom = styleableChild.getStyle("bottom");
							cVerticalCenter = styleableChild
								.getStyle("verticalCenter");

							if (usePercentSizing
									&& ((!isNaN(cTop)
											&& !isNaN(cBottom))
										|| (!isNaN(cTop)
											&& !isNaN(cVerticalCenter))
										|| (!isNaN(cBottom)
											&& !isNaN(cVerticalCenter))))
								usePercentSizing = false;
						}

						if (usePercentSizing)
							ph = ph * (100 / child.percentHeight);
						else
						{
							if (!isNaN(cTop) && !isNaN(cBottom))
								ph += cTop + cBottom;
							else if (!isNaN(cTop) && !isNaN(cVerticalCenter))
								ph += (cTop - cVerticalCenter) * 2;
							else if (!isNaN(cBottom) && !isNaN(cVerticalCenter))
								ph += (cBottom + cVerticalCenter) * 2;
							else if (!isNaN(cTop))
								ph += cTop;
							else if (!isNaN(cBottom))
								ph += cBottom;
							else if (!isNaN(cVerticalCenter))
								ph += Math.abs(cVerticalCenter) * 2;
						}
					}

					maxCH = Math.max(maxCH, Math.round(ph / child.gridHeight));
				}
			}

			if (isNaN(cw))
				cw = measuredCellWidth = maxCW;
			if (isNaN(ch))
				ch = measuredCellHeight = maxCH;
		}

		measuredWidth = cw * cc;
		measuredHeight = ch * rc;

		measuredWidth += vm.left + vm.right + hGap * (cc - 1);
		measuredHeight += vm.top + vm.bottom + vGap * (rc - 1);
	}

	override protected function updateDisplayList(unscaledWidth:Number,
			unscaledHeight:Number):void
	{
		super.updateDisplayList(unscaledWidth, unscaledHeight);

		var vm:EdgeMetrics = viewMetricsAndPadding;

		var hGap:Number = getStyle("horizontalGap");
		var vGap:Number = getStyle("verticalGap");

		var i:int;
		var child:ICanvasGridComponent;

		var cw:Number = isNaN(cellWidth) && isNaN(explicitWidth)
			? measuredCellWidth : cellWidth;
		var ch:Number = isNaN(cellHeight) && isNaN(explicitHeight)
			? measuredCellHeight : cellHeight;

		if (isNaN(cw))
		{
			var cc:int = columnCount;

			if (cc == -1)
			{
				var maxC:int = 0;

				for (i = 0; i < numChildren; i++)
				{
					child = getChildAt(i) as ICanvasGridComponent;
					maxC = Math.max(maxC, child.gridX + child.gridWidth);
				}

				cc = maxC;
			}

			cw = Math.round((unscaledWidth - vm.left - vm.right
						- hGap * (cc - 1)) / cc);
		}
		if (isNaN(ch))
		{
			var rc:int = rowCount;

			if (rc == -1)
			{
				var maxR:int = 0;

				for (i = 0; i < numChildren; i++)
				{
					child = getChildAt(i) as ICanvasGridComponent;
					maxR = Math.max(maxR, child.gridY + child.gridHeight);
				}

				rc = maxR;
			}

			ch = Math.round((unscaledHeight - vm.top - vm.bottom
						- vGap * (rc - 1)) / rc);
		}

		for (i = 0; i < numChildren; i++)
		{
			child = getChildAt(i) as ICanvasGridComponent;

			var dw:Number = child.gridWidth * cw
				+ (child.gridWidth - 1) * hGap;
			var dh:Number = child.gridHeight * ch
				+ (child.gridHeight - 1) * vGap;

			var posX:Number = vm.left + child.gridX * cw + child.gridX * hGap;
			var posY:Number = vm.top + child.gridY * ch + child.gridY * vGap;

			var w:Number = child.getExplicitOrMeasuredWidth();
			var h:Number = child.getExplicitOrMeasuredHeight();

			var styleableChild:IStyleClient = getChildAt(i) as IStyleClient;

			var cLeft:Number = NaN;
			var cRight:Number = NaN;
			var cHorizontalCenter:Number = NaN;

			var cTop:Number = NaN;
			var cBottom:Number = NaN;
			var cVerticalCenter:Number = NaN;

			if (styleableChild)
			{
				cLeft = styleableChild.getStyle("left");
				cRight = styleableChild.getStyle("right");
				cHorizontalCenter = styleableChild
					.getStyle("horizontalCenter");

				cTop = styleableChild.getStyle("top");
				cBottom = styleableChild.getStyle("bottom");
				cVerticalCenter = styleableChild
					.getStyle("verticalCenter");
			}

			if (!isNaN(cLeft) && !isNaN(cRight))
				w = dw - cLeft - cRight;
			else if (!isNaN(cLeft) && !isNaN(cHorizontalCenter))
				w = dw - (cLeft - cHorizontalCenter) * 2;
			else if (!isNaN(cRight) && !isNaN(cHorizontalCenter))
				w = dw - (cRight + cHorizontalCenter) * 2;
			else if (!isNaN(child.percentWidth))
				w = dw * (child.percentWidth / 100);

			if (!isNaN(cLeft))
				posX += cLeft;
			else if (!isNaN(cRight))
				posX += dw - cRight - w;
			else if (!isNaN(cHorizontalCenter))
				posX += (dw - w) / 2 + cHorizontalCenter;

			if (!isNaN(cTop) && !isNaN(cBottom))
				h = dh - cTop - cBottom;
			else if (!isNaN(cTop) && !isNaN(cVerticalCenter))
				h = dh - (cTop - cVerticalCenter) * 2;
			else if (!isNaN(cBottom) && !isNaN(cVerticalCenter))
				h = dh - (cBottom + cVerticalCenter) * 2;
			else if (!isNaN(child.percentWidth))
				h = dh * (child.percentWidth / 100);

			if (!isNaN(cTop))
				posY += cTop;
			else if (!isNaN(cBottom))
				posY += dh - cBottom - h;
			else if (!isNaN(cVerticalCenter))
				posY += (dh - h) / 2 + cVerticalCenter;

			child.move(posX, posY);
			child.setActualSize(w, h);
		}
	}
}

}

